"""Tests for rejester._queue

This software is released under an MIT/X11 open source license.

Copyright 2014 Diffeo, Inc.
"""

from __future__ import absolute_import

import logging
import os
import time

import pytest

from rejester.exceptions import ItemInUseError, LostLease

logger = logging.getLogger(__name__)
pytest_plugins = 'rejester.tests.fixtures'

def all_of_queue(queue):
    while True:
        item = queue.check_out_item(300)
        if item is None: return
        queue.return_item(item, None)
        yield item

def test_worker_id(rejester_queue):
    """simple worker_id test"""
    # since we start with a new namespace every time, the worker IDs
    # should start fresh too
    worker_id = rejester_queue.worker_id
    assert worker_id == 1

def test_one(rejester_queue):
    """Add 'a' with priority 1, and get it back"""
    rejester_queue.add_item('a', 1)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    rejester_queue.return_item(item, None)
    item = rejester_queue.check_out_item(300)
    assert item is None

def test_one_helper(rejester_queue):
    """Add 'a' with priority 1, and get it back"""
    rejester_queue.add_item('a', 1)
    assert list(all_of_queue(rejester_queue)) == ['a']

def test_interesting_priorities(rejester_queue):
    """Zero, one, many"""
    rejester_queue.add_item('a', 1)
    rejester_queue.add_item('b', 1048576)
    rejester_queue.add_item('c', -65536)
    rejester_queue.add_item('d', 0)
    rejester_queue.add_item('e', 0.5)
    rejester_queue.add_item('f', -0.1)
    rejester_queue.add_item('g', 0.2)
    assert list(all_of_queue(rejester_queue)) == ['b', 'a', 'e', 'g', 'd', 'f', 'c']

def test_add_checked_out(rejester_queue):
    """Adding a checked-out item should fail"""
    rejester_queue.add_item('a', 1)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    with pytest.raises(ItemInUseError):
        rejester_queue.add_item('a', 1)

def test_two(rejester_queue):
    """Add two items, out of order, and get them back"""
    rejester_queue.add_item('a', 2)
    rejester_queue.add_item('b', 1)
    assert list(all_of_queue(rejester_queue)) == ['a', 'b']
    rejester_queue.add_item('a', 1)
    rejester_queue.add_item('b', 2)
    assert list(all_of_queue(rejester_queue)) == ['b', 'a']

def test_reprioritize(rejester_queue):
    """Use add_item to change the priority of something"""
    rejester_queue.add_item('a', 1)
    rejester_queue.add_item('b', 2)
    rejester_queue.add_item('b', 0)
    assert list(all_of_queue(rejester_queue)) == ['a', 'b']

def test_requeue(rejester_queue):
    """Pull items out and put them back"""
    rejester_queue.add_item('a', 3)
    rejester_queue.add_item('b', 2)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    rejester_queue.return_item(item, 3)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    rejester_queue.return_item(item, 1)
    assert list(all_of_queue(rejester_queue)) == ['b', 'a']

def test_wrong_worker(rejester_queue, rejester_queue2):
    """If one queue instance checks out an item, another can't return it"""
    rejester_queue.add_item('a', 1)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    # NOTE: this also implicitly tests that creating a new queue
    # object gets a different worker ID
    with pytest.raises(LostLease):
        rejester_queue2.return_item(item, None)

def test_check_out_two(rejester_queue):
    """Nothing prevents the same worker from checking out two things"""
    rejester_queue.add_item('a', 2)
    rejester_queue.add_item('b', 1)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    item = rejester_queue.check_out_item(300)
    assert item == 'b'
    rejester_queue.return_item('a', None)
    rejester_queue.return_item('b', 1)
    assert list(all_of_queue(rejester_queue)) == ['b']

def test_reserve(rejester_queue, rejester_queue2):
    """Basic reservation/priority test

    1. Insert a@3, b@2
    2. Get, should return 'a'
    3. 'a' reserves 'b'
    4. Get, should be empty
    5. Return a@1
    6. Get, should return 'b'

    """
    rejester_queue.add_item('a', 3)
    rejester_queue.add_item('b', 2)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    reserved = rejester_queue.reserve_items(item, 'b')
    assert reserved == ['b']
    assert list(all_of_queue(rejester_queue2)) == []
    rejester_queue.return_item(item, 1)
    assert list(all_of_queue(rejester_queue2)) == ['b', 'a']

def test_reserve_twice(rejester_queue):
    """Reserving an item that's already reserved is a no-op"""
    rejester_queue.add_item('a', 2)
    rejester_queue.add_item('b', 1)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    reserved = rejester_queue.reserve_items(item, 'b')
    assert reserved == ['b']
    reserved = rejester_queue.reserve_items(item, 'b')
    assert reserved == []
    

def test_reserve_nonexistent(rejester_queue):
    """Reserving an item that doesn't exist is a no-op"""
    rejester_queue.add_item('a', 1)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    reserved = rejester_queue.reserve_items(item, 'b')
    assert reserved == []
    rejester_queue.return_item(item, None)

def test_reserve_self(rejester_queue):
    """Reserving yourself is a no-op"""
    rejester_queue.add_item('a', 1)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    reserved = rejester_queue.reserve_items(item, 'a')
    assert reserved == []
    rejester_queue.return_item(item, None)
    
def test_basic_expire(rejester_queue):
    """Basic expiration test"""
    rejester_queue.add_item('a', 1)
    item = rejester_queue.check_out_item(1)
    assert item == 'a'
    time.sleep(3)
    with pytest.raises(LostLease):
        rejester_queue.return_item(item, None)

def test_expiration_releases_reservations(rejester_queue, rejester_queue2):
    """If a reserves b, and a expires, then b should be released"""
    rejester_queue.add_item('a', 2)
    rejester_queue.add_item('b', 1)
    item = rejester_queue.check_out_item(1)
    assert item == 'a'
    reserved = rejester_queue.reserve_items(item, 'b')
    assert reserved == ['b']

    assert list(all_of_queue(rejester_queue2)) == []
    time.sleep(3)
    assert list(all_of_queue(rejester_queue2)) == ['a', 'b']

def test_renew(rejester_queue):
    """Basic renew test"""
    rejester_queue.add_item('a', 1)
    item = rejester_queue.check_out_item(2)
    assert item == 'a'
    time.sleep(1)
    rejester_queue.renew_item(item, 3)
    time.sleep(2)
    # at this point we have slept(3), which is after the original
    # expiration, so we should *not* get LostLease because we renewed
    rejester_queue.return_item(item, None)

def test_renew_overdue(rejester_queue):
    """Can't renew something that's already expired"""
    rejester_queue.add_item('a', 1)
    item = rejester_queue.check_out_item(1)
    assert item == 'a'
    time.sleep(3)
    with pytest.raises(LostLease):
        rejester_queue.renew_item(item, 3)

def test_dump(rejester_queue):
    """dump_queue() shouldn't crash; doesn't validate output"""
    rejester_queue.add_item('a', 1)
    rejester_queue.add_item('b', 0)
    item = rejester_queue.check_out_item(300)
    assert item == 'a'
    reserved = rejester_queue.reserve_items(item, 'b')
    assert reserved == ['b']
    rejester_queue.dump_queue('worker', 'available', 'priorities', 'expiration',
                              'workers', 'reservations_a', 'reservations_b',
                              'foo')



'''
rejester is a task manager written in python

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import argparse
import json
import lockfile
import logging
import os
from StringIO import StringIO
import sys
import termios
import traceback
import tty

import daemon
import yaml

from dblogger import configure_logging, FixedWidthFormatter
import rejester
from rejester._task_master import TaskMaster
from rejester.workers import run_worker, MultiWorker
import yakonfig

logger = logging.getLogger(__name__)

def stderr(m, newline='\n'):
    sys.stderr.write(m)
    sys.stderr.write(newline)
    sys.stderr.flush()

def getch():
    '''
    capture one char from stdin for responding to Y/N prompt
    '''
    fd = sys.stdin.fileno()
    if not os.isatty(fd):
        return sys.stdin.read(1)
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch

class MissingArgumentError(Exception):
    """Exception if more information needed to be given at the command line"""
    pass

class Manager(object):
    def __init__(self):
        self.config = yakonfig.get_global_config('rejester')
        self._task_master = None

    @property
    def task_master(self):
        """A `TaskMaster` object for manipulating work"""
        if self._task_master is None:
            self._task_master = TaskMaster(self.config)
        return self._task_master

    def _get_work_spec(self, **kwargs):
        work_spec_path  = kwargs.get('work_spec_path')
        if work_spec_path is None:
            raise MissingArgumentError('give a path to a work spec file '
                                       'with -w')
        if not os.path.exists(work_spec_path):
            raise MissingArgumentError('work spec file {!r} does not exist'
                                       .format(work_spec_path))
        work_spec = yaml.load(open(work_spec_path))
        return work_spec

    def _get_work_spec_name(self, **kwargs):
        name = kwargs.get('work_spec_name')
        if name is not None: return name
        if kwargs.get('work_spec_path') is None:
            raise MissingArgumentError('give a path to a work spec file '
                                       'with -w, or the work spec name '
                                       'with -W')
        work_spec = self._get_work_spec(**kwargs)
        return work_spec['name']

    def load(self, **kwargs):
        '''loads work_units into a namespace for a given work_spec
        '''
        work_spec = self._get_work_spec(**kwargs)

        work_units_path = kwargs.get('work_units_path')
        if work_units_path is None:
            raise MissingArgumentError('give a path to a work unit file '
                                       'with -u')
        if work_units_path == '-':
            work_units_fh = sys.stdin
        elif work_units_path.endswith('.gz'):
            work_units_fh = gzip.open(work_units_path)
        else:
            work_units_fh = open(work_units_path)
        print 'loading work units from %r' % work_units_fh
        work_units = dict()
        for line in work_units_fh:
            work_unit = json.loads(line)
            work_units.update(work_unit)
        print 'pushing work units'
        self.task_master.update_bundle(work_spec, work_units)
        print 'finished writing %d work units to work_spec=%r' \
            % (len(work_units), work_spec['name'])

    def delete(self, **kwargs):
        'make sure the user means to delete'
        namespace = self.config['namespace']
        stderr('Delete everything in %r?  Enter namespace: ' % namespace, newline='')
        if kwargs.get('assume_yes', False):
            stderr('... assuming yes.\n')
            do_delete = True
        else:
            idx = 0
            assert len(namespace) > 0
            while idx < len(namespace):
                ch = getch()
                if ch == namespace[idx]:
                    idx += 1
                    do_delete = True
                else:
                    do_delete = False
                    break

        if do_delete:
            stderr('\nDeleting ...')
            sys.stdout.flush()
            self.task_master.registry.delete_namespace()
            stderr('')

        else:
            stderr(' ... Aborting.')

    def status(self, **kwargs):
        work_spec_name = self._get_work_spec_name(**kwargs)
        print json.dumps(self.task_master.status(work_spec_name), indent=4, sort_keys=True)

    def work_units(self, **kwargs):
        work_spec_name = self._get_work_spec_name(**kwargs)
        work_units = self.task_master.list_work_units(work_spec_name)
        for k in sorted(work_units.keys()):
            if kwargs.get('verbose'):
                print '{!r}: {!r}'.format(k, work_units[k])
            else:
                print k

    def set_IDLE(self, **kwargs):
        self.task_master.set_mode(self.task_master.IDLE)
        print 'set mode to IDLE'

    def set_RUN(self, **kwargs):
        self.task_master.set_mode(self.task_master.RUN)
        print 'set mode to RUN'

    def set_TERMINATE(self, **kwargs):
        self.task_master.set_mode(self.task_master.TERMINATE)
        print 'set mode to TERMINATE'

    def run_worker(self, **kwargs):
        pidfile = kwargs.get('pidfile')
        if pidfile is not None:
            if not os.path.isabs(pidfile):
                raise MissingArgumentError('--pidfile requires an '
                                           'absolute path')
            if not os.path.exists(os.path.dirname(pidfile)):
                raise MissingArgumentError('--pidfile path {!r} does not exist'
                                           .format(os.path.dirname(pidfile)))

        logpath = kwargs.get('logpath')
        if logpath is not None:
            if not os.path.isabs(logpath):
                raise MissingArgumentError('--logpath requires an '
                                           'absolute path')
            if not os.path.exists(os.path.dirname(logpath)):
                raise MissingArgumentError('--logpath path {!r} does not exist'
                                           .format(os.path.dirname(logpath)))

        if pidfile:
            pidfile_lock = lockfile.FileLock(pidfile)
        else:
            pidfile_lock = None
        context = daemon.DaemonContext(pidfile=pidfile_lock)
        with context:
            try:
                if pidfile:
                    open(pidfile,'w').write(str(os.getpid()))
                # Holding loggers open across DaemonContext is a big
                # problem; establish them for the first time here
                configure_logging(yakonfig.get_global_config())
                if logpath:
                    formatter = FixedWidthFormatter()
                    # TODO: do we want byte-size RotatingFileHandler or TimedRotatingFileHandler?
                    handler = logging.handlers.RotatingFileHandler(
                        logpath, maxBytes=10000000, backupCount=3)
                    handler.setFormatter(formatter)
                    logging.getLogger('').addHandler(handler)
                logger.debug('inside daemon context')
                run_worker(MultiWorker, self.config)
                logger.debug('run_worker exited')
            except Exception, exc:
                logp = logpath or os.path.join('/tmp', 'rejester-failure.log')
                open(logp, 'a').write(traceback.format_exc(exc))
                raise


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('action', help='must be one of: load, delete, status, run')
    parser.add_argument('--pidfile', default=None, help='PID lock file for use with action=run')
    parser.add_argument('-y', '--yes', default=False, action='store_true', dest='assume_yes',
                        help='Assume "yes" and require no input for confirmation questions.')
    parser.add_argument('-w', '--work-spec', dest='work_spec_path',
                        help='path to a YAML or JSON file containing a Work Spec')
    parser.add_argument('-W', '--work-spec-name',
                        help='name of a work spec for queries')
    parser.add_argument('-u', '--work-units', default='-', dest='work_units_path',
                        help='path to file with one JSON record per line, each describing a Work Unit')
    parser.add_argument('--logpath', default=None)
    parser.add_argument('-v', '--verbose', default=False, action='store_true',
                        help='include more information in output')
    args = yakonfig.parse_args(parser, [yakonfig, rejester])

    # Split actions by comma, and execute them in sequence
    actions = args.action.split(',')
    for action_string in actions:
        if action_string not in Manager.__dict__:
            parser.error('Unrecognized action "{}"'.format(action_string))

    # run_worker is Very Special; anything else needs to set up logging now
    if 'run_worker' not in actions:
        configure_logging(yakonfig.get_global_config())

    mgr = Manager()

    for action_string in actions:
        action = getattr(mgr, action_string)
        try:
            action(**args.__dict__)
        except MissingArgumentError, e:
            parser.error('{}: {!s}'.format(action_string, e))

if __name__ == '__main__':
    main()



'''Rejester workers.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2013 Diffeo, Inc.
'''
from __future__ import absolute_import
import os
import sys
import time
import uuid
import logging
import gevent
import psutil
import random
import Queue
import multiprocessing
from signal import signal, SIGHUP, SIGTERM, SIGABRT
from operator import itemgetter
from collections import deque

from rejester import TaskMaster, Worker
from rejester._task_master import WORKER_OBSERVED_MODE, WORKER_STATE_

logger = logging.getLogger(__name__)

def test_work_program(work_unit):
    ## just to show that this works, we get the config from the data
    ## and *reconnect* to the registry with a second instances instead
    ## of using work_unit.registry
    config = work_unit.data['config']
    sleeptime = float(work_unit.data.get('sleep', 9.0))
    task_master = TaskMaster(config)
    logger.info('executing work_unit %r ... %s', work_unit.key, sleeptime)
    time.sleep(sleeptime)  # pretend to work
    logger.info('finished %r' % work_unit)

def run_worker(worker_class, *args, **kwargs):
    '''Bridge function to run a worker under :mod:`multiprocessing`.

    The :mod:`multiprocessing` module cannot
    :func:`~multiprocessing.apply_async` to a class constructor, even
    if the ``__init__`` calls ``.run()``, so this simple wrapper calls
    ``worker_class(*args, **kwargs)`` and logs any exceptions before
    re-raising them.

    '''
    try:
        worker = worker_class(*args, **kwargs)
    except Exception, exc:
        logger.critical('failed to create worker {!r}'.format(worker_class),
                        exc_info=True)
        raise
    try:
        worker.register()
        worker.run()
    except Exception, exc:
        logger.error('worker {!r} died'.format(worker_class), exc_info=True)
        raise
    finally:
        worker.unregister()


class HeadlessWorker(Worker):
    '''This expects to receive a WorkUnit from its parent process, which
    is running MultiWorker.
    '''

    def __init__(self, config, worker_id, work_spec_name, work_unit_key):
        super(HeadlessWorker, self).__init__(config)
        for sig_num in [SIGTERM, SIGHUP, SIGABRT]:
            signal(sig_num, self.terminate)
        self.work_unit = self.task_master.get_assigned_work_unit(
            worker_id, work_spec_name, work_unit_key)
        ## carry this to overwrite self.worker_id after .register()
        self._pre_assigned_worker_id = worker_id

    def register(self):
        super(HeadlessWorker, self).register()
        self.worker_id = self._pre_assigned_worker_id

    def run(self):
        self.work_unit.run()

    def terminate(self, sig_num, frame):
        logger.info('received %d, ending work unit', sig_num)
        self.work_unit.terminate()
        sys.exit()

class MultiWorker(Worker):
    '''launches one child process per core on the machine, and reports
    available_gb based on what measurements.  This class manages the
    TaskMaster interactions and sends WorkUnit instances to child
    processes.
    '''
    def __init__(self, config):
        super(MultiWorker, self).__init__(config)
        self._event_queue = multiprocessing.Queue()
        self._mode = None
        self.pool = None

    _available_gb = None

    @classmethod
    def available_gb(cls):
        if cls._available_gb is None:
            mem = psutil.phymem_usage()
            cls._available_gb = float(mem.free) / multiprocessing.cpu_count()
        return cls._available_gb

    def _finish_callback(self, *args):
        # We don't actually get anything useful from the work call, so
        # just post an event that causes us to wake up and poll all
        # the slots.
        self._event_queue.put(True)

    def _poll_async_result(self, async_result, work_unit, do_update=True):
        if async_result is None:
            return
        assert work_unit is not None
        if not async_result.ready():
            if do_update:
                logger.debug('not ready %r, update', work_unit.key)
                work_unit.update()
            return
        try:
            async_result.get(0)
        except multiprocessing.TimeoutError:
            if do_update:
                logger.debug('get timeout update %r', work_unit.key)
                work_unit.update()
            return
        except Exception, exc:
            logger.error('trapped child exception', exc_info=True)
            work_unit.fail(exc)
        else:
            ## if it gets here, slot should always be finished
            assert async_result.ready()
            work_unit.finish()
        ## either failed or finished
        assert work_unit.failed or work_unit.finished

    def _get_and_start_work(self):
        "return (async_result, work_unit) or (None, None)"
        worker_id = uuid.uuid4().hex
        work_unit = self.task_master.get_work(worker_id, available_gb=self.available_gb())
        if work_unit is None:
            return None, None
        async_result = self.pool.apply_async(
            run_worker,
            (HeadlessWorker, self.task_master.registry.config,
             worker_id,
             work_unit.work_spec_name,
             work_unit.key),
            callback=self._finish_callback)
        return async_result, work_unit

    def _poll_slots(self, slots, mode=None, do_update=False):
        hasWork = True
        for i in xrange(len(slots)):
            async_result, work_unit = slots[i]
            if (async_result is not None) and (work_unit is not None):
                self._poll_async_result(async_result, work_unit, do_update=do_update)
                if work_unit.failed or work_unit.finished:
                    slots[i] = (None, None)
            if (slots[i][0] is None) and (mode == self.task_master.RUN):
                if hasWork:
                    slots[i] = self._get_and_start_work()
                    if slots[i][0] is None:
                        # If we fail to get work, don't hammer the
                        # taskmaster with requests for work. Wait
                        # until after a sleep and the next _poll_slots
                        # cycle.
                        hasWork = False

    def run(self):
        tm = self.task_master
        num_workers = multiprocessing.cpu_count()
        if self.pool is None:
            self.pool = multiprocessing.Pool(num_workers, maxtasksperchild=1)
        ## slots is a fixed-length list of [AsyncRsults, WorkUnit]
        slots = [[None, None]] * num_workers
        logger.info('MultiWorker starting with %s workers', num_workers)
        min_loop_time = 2.0
        lastFullPoll = time.time()
        while True:
            mode = self.heartbeat()
            if mode != self._mode:
                logger.info('worker {} changed to mode {}'
                            .format(self.worker_id, mode))
                self._mode = mode
            now = time.time()
            should_update = (now - lastFullPoll) > min_loop_time
            self._poll_slots(slots, mode=mode, do_update=should_update)
            if should_update:
                lastFullPoll = now

            if mode == tm.TERMINATE:
                num_waiting = sum(map(int, map(bool, map(itemgetter(0), slots))))
                if num_waiting == 0:
                    logger.info('MultiWorker all children finished')
                    break
                else:
                    logger.info('MultiWorker waiting for %d children to finish', num_waiting)

            sleepsecs = random.uniform(1,5)
            sleepstart = time.time()
            try:
                self._event_queue.get(block=True, timeout=sleepsecs)
                logger.debug('woken by event looptime=%s sleeptime=%s', sleepstart - now, time.time() - sleepstart)
            except Queue.Empty, empty:
                logger.debug('queue timed out. be exhausting, looptime=%s sleeptime=%s', sleepstart - now, time.time() - sleepstart)
                # it's cool, timed out, do the loop of checks and stuff.

        logger.info('MultiWorker exiting')



'''
http://github.com/diffeo/rejester

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import time
import uuid
import json
import logging
import random
import socket
import atexit
import contextlib
from uuid import UUID
from functools import wraps
from collections import defaultdict

import redis

from rejester.exceptions import EnvironmentError, LockError, \
    ProgrammerError
from rejester._redis import RedisBase

logger = logging.getLogger(__name__)

class Registry(RedisBase):
    '''Store string-keyed dictionaries in Redis.

    Provides a centralized storage mechanism for dictionaries,
    including atomic operations for moving (key, value) pairs between
    dictionaries, and incrementing counts.

    Most operations on the registry require getting a lock via
    the database, for instance

    >>> with registry.lock() as session:
    ...   value = session.get(k1, k2)

    The lock mechanism ensures that no two Registry objects do work
    concurrently, even running on separate systems.

    The basic data object is a string-keyed dictionary stored under
    some key.  The dictionary keys are also stored in a prioritized
    list.  This in effect provides two levels of dictionary, using the
    Redis key and the dictionary key.  The registry makes an effort to
    store all types of object as values, potentially serializing them
    into JSON.  Core operations include

    `update(key, {'k': 'v'})` adds items to a dictionary

    `popmany(key, 'k1', 'k2')` removes values from a dictionary

    `len(key)` returns the number of items in a dictionary

    `getitem_reset(key)` returns the highest-priority item, leaving it
    in the queue with a new priority

    `popitem(key)` returns the highest-priority item, removing it

    `popitem_move(key1, key2)` moves the highest-priority item from
    `key1` to `key2`

    `move(key1, key2, {'k': anything})` moves values between dictionaries

    `move_all(key1, key2)` moves everything from one dictionary to another

    `pull(key1)` gets the entire contents of a dictionary

    `increment(key, 'k')` gets an increasing numeric value

    `get(key, 'k')`, `set(key, 'k', 'v')`, `delete(key)` do what they
    sound like

    '''

    # This makes extensive use of Lua scripts in Redis; see
    # http://redis.io/commands/eval for basics.  This provides atomicity
    # (Redis never runs anything in parallel with a Lua script), but
    # also means that much of the logic is in a hard-to-watch script,
    # and that performance of these scripts is pretty critical.

    def __init__(self, config):
        '''
        Initialize the registry using a config
        '''
        super(Registry, self).__init__(config)

        ## populated only when lock is acquired
        self._lock_name = None
        self._session_lock_identifier = None

        self._startup()
        atexit.register(self._exit)

        logger.debug('worker_id=%r  starting up on hostname=%r', 
                     self.worker_id, socket.gethostbyname(socket.gethostname()))

    def _startup(self):
        conn = redis.Redis(connection_pool=self.pool)

        ## Get a unique worker id from the registry
        self.worker_id = conn.incrby(self._namespace('worker_ids'), 1)

        ## Add worker id to list of current workers
        conn.sadd(self._namespace('cur_workers'), self.worker_id)

        ## store the local_ip of this worker for debugging
        conn.set(self._namespace(self.worker_id) + '_ip', self._local_ip)

    def _exit(self):
        conn = redis.Redis(connection_pool=self.pool)

        ## cleanup the worker's state
        #with self.lock() as session:
        #    session...

        ## Remove the worker from the list of workers
        conn.srem(self._namespace('cur_workers'), self.worker_id)
        logger.debug('atexit: worker_id=%r closed registry client', self.worker_id)

    def _conn(self):
        '''debugging aid to easily grab a connection from the connection pool

        '''
        conn = redis.Redis(connection_pool=self.pool)
        return conn

    def _all_workers(self):
        conn = redis.Redis(connection_pool=self.pool)
        return conn.smembers(self._namespace('cur_workers'))

    def _acquire_lock(self, lock_name, identifier, atime=600, ltime=600):
        '''Acquire a lock for a given identifier.  Contend for the lock for
        atime, and hold the lock for ltime.
        '''
        conn = redis.Redis(connection_pool=self.pool)
        end = time.time() + atime
        while end > time.time():
            if conn.set(lock_name, identifier, ex=ltime, nx=True):
                logger.debug("won lock %s" % lock_name)
                return identifier
            sleep_time = random.uniform(0, 3)
            time.sleep(sleep_time)

        logger.warn('failed to acquire lock %s for %f seconds', lock_name, atime)
        return False


    def re_acquire_lock(self, ltime=600):
        '''Re-acquire the lock'''
        conn = redis.Redis(connection_pool=self.pool)
        if conn.set(self._lock_name, self._session_lock_identifier, ex=ltime, xx=True):
            logger.debug('re-acquired lock %s', self._lock_name)
            return self._session_lock_identifier

        raise EnvironmentError('failed to re-acquire lock')


    def _release_lock(self, lock_name, identifier):
        '''This follows a 'test and delete' pattern.  See redis documentation
        http://redis.io/commands/set

        This is needed because the lock could have expired before we
        deleted it.  We can only delete the lock if we still own it.
        '''
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            return redis.call("del", KEYS[1])
        else
            return -1
        end
        '''
        conn = redis.Redis(connection_pool=self.pool)
        num_keys_deleted = conn.eval(script, 1, lock_name, identifier)
        if num_keys_deleted == -1:
            ## registry_lock no long owned by this process
            raise EnvironmentError(
                'Lost lock in registry already')
        return True

    @contextlib.contextmanager
    def lock(self, lock_name='global_registry_lock', atime=600, ltime=600):
        '''
        A context manager wrapper around acquring a lock
        '''
        ## Prepend namespace to lock_name
        lock_name = self._namespace(lock_name)

        identifier = str(uuid.uuid4())
        if self._acquire_lock(lock_name, identifier, atime,
                              ltime) != identifier:
            raise LockError("could not acquire lock")
        try:
            self._lock_name = lock_name
            self._session_lock_identifier = identifier
            yield self
        finally:
            self._release_lock(lock_name, identifier)
            self._lock_name = None
            self._session_lock_identifier = None

    def _encode(self, data):
        '''Redis hash's store strings in the keys and values.  Since we want
        to store UUIDs and tuples of UUIDs, we cannot just use json.
        Instead, we employ a very simply decoding to turn various
        strings into datatypes we understand.
        '''
        if isinstance(data, tuple):
            return 't:' + '-'.join([self._encode(item) for item in data])
        elif isinstance(data, UUID):
            return 'u:' + str(data.int)
        else:
            try:
                return 'j:' + json.dumps(data)
            except Exception, exc:
                logger.error('Fail to encode data %r', data, exc_info=True)
                raise TypeError

    def _decode(self, string):
        '''Redis hash's store strings in the keys and values.  Since we want
        to store UUIDs and tuples of UUIDs, we cannot just use json.
        Instead, we employ a very simply decoding to turn various
        strings into datatypes we understand.

        Note, that redis would convert things into a string
        representation that we could 'eval' here, but it is not a good
        practice from a security standpoint to eval data from a
        database.
        '''
        if len(string) == 0:
            return None
        if string[0] == 't':
            return tuple([self._decode(item) for item in
                          string[2:].split('-')])
        elif string[0] == 'u':
            return UUID(int=int(string[2:]))
        elif string[0] == 'j':
            try:
                if string[2:]:
                    return json.loads(string[2:])
                else:
                    None
            except ValueError, exc:
                logger.critical('%r --> tried to json.loads(%r)', string, string[2:], exc_info=True)
                raise
        else:
            try:
                return float(string)
            except:
                pass
            ## Raise a type error on all other types of data
            logger.error('Fail to decode string %r', string, exc_info=True)
            raise TypeError

    def update(self, dict_name, mapping=None, priorities=None, expire=None,
               locks=None):
        '''Add mapping to a dictionary, replacing previous values

        Can be called with only dict_name and expire to refresh the
        expiration time.

        NB: locks are only enforced if present, so nothing prevents
        another caller from coming in an modifying data without using
        locks.

        :param mapping: a dict of keys and values to update in
          dict_name.  Must be specified if priorities is specified.
        :param priorities: a dict with the same keys as those in
          mapping that provides a numerical value indicating the
          priority to assign to that key.  Default sets 0 for all keys.
        :param int expire: if specified, then dict_name will be set to
          expire in that many seconds.
        :param locks: a dict with the same keys as those in the
          mapping.  Before making any particular update, this function
          checks if a key is present in a 'locks' table for this dict,
          and if so, then its value must match the value provided in the
          input locks dict for that key.  If not, then the value
          provided in the locks dict is inserted into the 'locks' table.
          If the locks parameter is None, then no lock checking is
          performed.
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        ## script is evaluated with numkeys=2, so KEYS[1] is lock_name,
        ## KEYS[2] is dict_name, ARGV[1] is identifier, and ARGV[i>=2]
        ## are keys, values, and priorities arranged in triples.
        if priorities is None:
            ## set all priorities to zero
            priorities = defaultdict(int)
        if locks is None:
            ## set all locks to None
            locks = defaultdict(lambda: '')
        if not (expire is None or isinstance(expire, int)):
            raise ProgrammerError('expire must be int or unspecified')
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            for i = 3, #ARGV, 4  do
                if ARGV[i+3] ~= 'j:""' then
                    local curr_lock = redis.call("hget",  KEYS[2] .. "_locks", ARGV[i])                    
                    if curr_lock and curr_lock ~= ARGV[i+3] then
                        return {-1, ARGV[i], curr_lock, ARGV[i+3]}
                    end
                    redis.call("hset",  KEYS[2] .. "_locks", ARGV[i], ARGV[i+3])
                end
            end
            for i = 3, #ARGV, 4  do
                redis.call("hset",  KEYS[2], ARGV[i], ARGV[i+1])
                redis.call("zadd",  KEYS[2] .. "keys", ARGV[i+2], ARGV[i])
            end
            if tonumber(ARGV[2]) ~= nil then
                redis.call("expire", KEYS[2], ARGV[2])
                redis.call("expire", KEYS[2] .. "keys", ARGV[2])
            end
            return {1, 0}
        else
            -- ERROR: No longer own the lock
            return {0, 0}
        end
        '''
        dict_name = self._namespace(dict_name)
        if mapping is None:
            mapping = {}
        items = []
        ## This flattens the dictionary into a list
        for key, value in mapping.iteritems():
            items.append(self._encode(key))
            items.append(self._encode(value))
            items.append(priorities[key])
            items.append(self._encode(locks[key]))

        conn = redis.Redis(connection_pool=self.pool)
        #logger.debug('update %r %r', dict_name, items)
        res = conn.eval(script, 2, self._lock_name, dict_name, 
                        self._session_lock_identifier, expire, 
                        *items)
        if res[0] == 0:
            raise EnvironmentError(
                'Unable to add items to %s in registry' % dict_name)
        elif res[0] == -1:
            raise EnvironmentError(
                'lost lock on key=%r owned by %r not %r in %s' 
                % (self._decode(res[1]), res[2], res[3], dict_name))


    def reset_priorities(self, dict_name, priority):
        '''set all priorities in dict_name to priority

        :type priority: float or int
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        ## see comment above for script in update
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            local keys = redis.call('ZRANGE', KEYS[2] .. "keys", 0, -1)
            for i, next_key in ipairs(keys) do
                redis.call("zadd",  KEYS[2] .. "keys", ARGV[2], next_key)
            end
            return 1
        else
            -- ERROR: No longer own the lock
            return 0
        end
        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        res = conn.eval(script, 2, self._lock_name, dict_name, self._session_lock_identifier, priority)
        if not res:
            # We either lost the lock or something else went wrong
            raise EnvironmentError(
                'Unable to add items to %s in registry' % dict_name)


    def popmany(self, dict_name, *keys):
        '''
        map(D.pop, keys), remove specified keys
        If key is not found, do nothing

        :returns None:
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        ## see comment above for script in update
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            for i = 2, #ARGV do
                redis.call("hdel", KEYS[2], ARGV[i])
                redis.call("zrem", KEYS[2] .. "keys", ARGV[i])
            end
            return 1
        else
            -- ERROR: No longer own the lock
            return 0
        end
        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        encoded_keys = [self._encode(key) for key in keys]
        res = conn.eval(script, 2, self._lock_name, dict_name,
                        self._session_lock_identifier, *encoded_keys)        
        if len(keys) > 0 and res == 0:
            # We either lost the lock or something else went wrong
            raise EnvironmentError(
                'Unable to remove items from %s in registry: %r' 
                % (dict_name, res))

    def len(self, dict_name, priority_min='-inf', priority_max='+inf'):
        '''returns number of items in dict_name within
        [priority_min, priority_max]
        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        return conn.zcount(dict_name + 'keys', priority_min, priority_max)

    def getitem_reset(
            self, dict_name, priority_min='-inf', priority_max='+inf',
            new_priority=0,
            lock=None
        ):
        '''D.getitem_reset() -> (key, value), return some pair that has a
        priority that satisfies the priority_min/max constraint; but
        return None if that range is empty.

        (key, value) stays in the dictionary and is assigned new_priority
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        ## see comment above for script in update
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            -- remove next item of dict_name
            local next_key, next_priority = redis.call("zrangebyscore", KEYS[2] .. "keys", ARGV[2], ARGV[3], "WITHSCORES")[1]

            if not next_key then
                return {}
            end

            local next_val = redis.call("hget", KEYS[2], next_key)
            redis.call("zadd", KEYS[2] .. "keys", ARGV[4], next_key)
            if ARGV[5] then
                redis.call("hset", KEYS[2] .. "_locks", next_key, ARGV[5])
                redis.call("hset", KEYS[2] .. "_locks", ARGV[5], next_key)
            else
                local old_key2 = redis.call("hget", KEYS[2] .. "_locks", next_key)
                redis.call("hdel", KEYS[2] .. "_locks", next_key)
                redis.call("hdel", KEYS[2] .. "_locks", old_key2)
            end

            return {next_key, next_val}
        else
            -- ERROR: No longer own the lock
            return -1
        end
        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        logger.debug('getitem_reset: %s %s %s',
                     self._lock_name, self._session_lock_identifier, dict_name)
        key_value = conn.eval(script, 2, self._lock_name, dict_name, 
                              self._session_lock_identifier,
                              priority_min, priority_max,
                              new_priority,
                              self._encode(lock),
                              )
        if key_value == -1:
            raise KeyError(
                'Registry failed to return an item from %s' % dict_name)

        if key_value == []:
            return None

        return self._decode(key_value[0]), self._decode(key_value[1])

    def popitem(self, dict_name, priority_min='-inf', priority_max='+inf'):
        '''
        D.popitem() -> (k, v), remove and return some (key, value) pair as a
        2-tuple; but raise KeyError if D is empty.
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        ## see comment above for script in update
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            -- remove next item of dict_name
            local next_key, next_priority = redis.call("zrangebyscore", KEYS[2] .. "keys", ARGV[2], ARGV[3], "WITHSCORES")[1]

            if not next_key then
                return {}
            end
            
            redis.call("zrem", KEYS[2] .. "keys", next_key)
            local next_val = redis.call("hget", KEYS[2], next_key)
            -- zrem removed it from list, so also remove from hash
            redis.call("hdel", KEYS[2], next_key)
            return {next_key, next_val}
        else
            -- ERROR: No longer own the lock
            return -1
        end
        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        logger.critical('popitem: %s %s %s' 
                        % (self._lock_name, self._session_lock_identifier, dict_name))
        key_value = conn.eval(script, 2, self._lock_name, dict_name, 
                              self._session_lock_identifier,
                              priority_min, priority_max,
                              )
        if key_value == -1:
            raise KeyError(
                'Registry failed to return an item from %s' % dict_name)

        if key_value == []:
            return None

        return self._decode(key_value[0]), self._decode(key_value[1])

    def popitem_move(self, from_dict, to_dict, priority_min='-inf', priority_max='+inf'):
        '''
        Pop an item out of from_dict, store it in to_dict, and return it
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            -- find the next key and priority
            local next_items = redis.call("zrangebyscore", KEYS[2] .. "keys", ARGV[2], ARGV[3], "WITHSCORES")
            local next_key = next_items[1]
            local next_priority = next_items[2]
            
            if not next_key then
                return {}
            end

            -- remove next item of from_dict
            redis.call("zrem", KEYS[2] .. "keys", next_key)

            local next_val = redis.call("hget", KEYS[2], next_key)
            -- zrem removed it from list, so also remove from hash
            redis.call("hdel", KEYS[2], next_key)

            -- put it in to_dict
            redis.call("hset", KEYS[3], next_key, next_val)
            redis.call("zadd", KEYS[3] .. "keys", next_priority, next_key)

            return {next_key, next_val, next_priority}
        else
            -- ERROR: No longer own the lock
            return -1
        end
        '''
        conn = redis.Redis(connection_pool=self.pool)
        key_value = conn.eval(script, 3, self._lock_name, 
                              self._namespace(from_dict), 
                              self._namespace(to_dict), 
                              self._session_lock_identifier,
                              priority_min, priority_max,
                              )

        if key_value == []:
            return None

        if None in key_value or key_value == -1:
            raise KeyError(
                'Registry.popitem_move(%r, %r) --> %r' % (from_dict, to_dict, key_value))

        logger.debug('Registry.popitem_move(%r, %r) --> %r', from_dict, to_dict, key_value)
        return self._decode(key_value[0]), self._decode(key_value[1])

    def move(self, from_dict, to_dict, mapping):
        '''Pop mapping's keys out of from_dict, and update them in to_dict
        with the new values provided by mapping
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            local count = 0
            for i = 2, #ARGV, 2  do
                -- remove next item of from_dict
                local next_priority = redis.call("zscore", KEYS[2] .. "keys", ARGV[i])
                redis.call("zrem", KEYS[2] .. "keys", ARGV[i])
                -- drop the old value: local next_val = redis.call("hget", KEYS[2], ARGV[i])
                -- zrem removed it from Sorted Set, so also remove from hash
                redis.call("hdel", KEYS[2], ARGV[i])
                -- put it in to_dict
                redis.call("hset",  KEYS[3], ARGV[i], ARGV[i+1])
                redis.call("zadd", KEYS[3] .. "keys", next_priority, ARGV[i])
                count = count + 1
            end
            return count
        else
            -- ERROR: No longer own the lock
            return 0
        end
        '''
        items = []
        ## This flattens the dictionary into a list
        for key, value in mapping.iteritems():
            key = self._encode(key)
            items.append(key)
            value = self._encode(value)
            items.append(value)

        conn = redis.Redis(connection_pool=self.pool)

        #logger.debug('move %r %r %r', from_dict, to_dict, items)
        num_moved = conn.eval(script, 3, self._lock_name, 
                              self._namespace(from_dict), 
                              self._namespace(to_dict), 
                              self._session_lock_identifier, *items)
        if num_moved != len(items) / 2:
            raise EnvironmentError(
                'Registry failed to move all: num_moved = %d != %d len(items)'
                % (num_moved, len(items)))


    def move_all(self, from_dict, to_dict):
        '''
        move all items out of from_dict and update them in to_dict
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            local count = 0
            local keys = redis.call('ZRANGE', KEYS[2] .. "keys", 0, -1)
            for i, next_key in ipairs(keys) do
                -- get the value and priority for this key
                local next_val = redis.call("hget", KEYS[2], next_key)
                local next_priority = redis.call("zscore", KEYS[2] .. "keys", next_key)
                -- remove item of from_dict
                redis.call("zrem", KEYS[2] .. "keys", next_key)
                -- also remove from hash
                redis.call("hdel", KEYS[2], next_key)
                -- put it in to_dict
                redis.call("hset",  KEYS[3], next_key, next_val)
                redis.call("zadd", KEYS[3] .. "keys", next_priority, next_key)
                count = count + 1
            end
            return count
        else
            -- ERROR: No longer own the lock
            return 0
        end
        '''
        conn = redis.Redis(connection_pool=self.pool)

        num_moved = conn.eval(script, 3, self._lock_name, 
                              self._namespace(from_dict), 
                              self._namespace(to_dict), 
                              self._session_lock_identifier)
        return num_moved


    def pull(self, dict_name):
        '''
        Pull entire dictionary
        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        res = conn.hgetall(dict_name)
        split_res = {self._decode(key):
                     self._decode(value)
                     for key, value in res.iteritems()}
        return split_res

    def filter(self, dict_name, priority_min='-inf', priority_max='+inf'):
        '''Pull all keys and values in dictionary between priority_min/max
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            -- find all the keys and priorities within range
            local next_keys = redis.call("zrangebyscore", KEYS[2] .. "keys", ARGV[2], ARGV[3])
            
            if not next_keys[1] then
                return {}
            end

            local t = {}
            for i = 1, #next_keys  do
                local next_val = redis.call("hget", KEYS[2], next_keys[i])
                table.insert(t, next_keys[i])
                table.insert(t, next_val)
            end

            return t
        else
            -- ERROR: No longer own the lock
            return -1
        end
        '''
        conn = redis.Redis(connection_pool=self.pool)
        res = conn.eval(script, 2, self._lock_name, 
                        self._namespace(dict_name), 
                        self._session_lock_identifier,
                        priority_min, priority_max,
        )
        if None in res or res == -1:
            raise KeyError(
                'Registry.filter(%r, %r, %r) --> %r' 
                % (dict_name, priority_min, priority_max, res))

        logger.debug('Registry.filter(%r, %r, %r) --> %r', 
                     dict_name, priority_min, priority_max, res)
        split_res = {self._decode(res[i]):
                     self._decode(res[i+1])
                     for i in xrange(0, len(res)-1, 2)}
        return split_res

    def set_1to1(self, dict_name, key1, key2):
        '''set two keys to be equal in a 1-to-1 mapping.
        '''
        if self._lock_name is None:
            raise ProgrammerError('must acquire lock first')
        script = '''
        if redis.call("get", KEYS[1]) == ARGV[1]
        then
            redis.call("hset", KEYS[2], ARGV[2], ARGV[3])
            redis.call("hset", KEYS[2], ARGV[3], ARGV[2])
        else
            -- ERROR: No longer own the lock
            return -1
        end
        '''
        conn = redis.Redis(connection_pool=self.pool)
        res = conn.eval(script, 2, self._lock_name, 
                        self._namespace(dict_name), 
                        self._session_lock_identifier,
                        self._encode(key1),
                        self._encode(key2),
        )
        if res == -1:
            raise EnvironmentError()

    def get(self, dict_name, key, default=None, include_priority=False):
        '''
        get value for key, if missing return default if provided
        '''
        dict_name = self._namespace(dict_name)
        key = self._encode(key)
        conn = redis.Redis(connection_pool=self.pool)
        val = conn.hget(dict_name, key)
        _val = val and self._decode(val) or default
        if include_priority:
            if val:
                priority = conn.zscore(dict_name + 'keys', key)
                return _val, priority
            else:
                return _val, None
        return _val

    def set(self, dict_name, key, value, priority=None):
        '''
        set value for key
        '''
        if priority is not None:
            priorities = {key: priority}
        else:
            priorities = None
        self.update(dict_name, {key: value}, priorities=priorities)

    def delete(self, dict_name):
        '''
        delete the entire hash named dict_name
        '''
        dict_name = self._namespace(dict_name)
        conn = redis.Redis(connection_pool=self.pool)
        conn.delete(dict_name, dict_name + 'keys')

    def direct_call(self, *args):
        '''execute a direct redis call against this Registry instances
        namespaced keys.  This is low level is should only be used for
        prototyping.

        arg[0] = redis function
        arg[1] = key --- will be namespaced before execution
        args[2:] = args to function

        :returns raw return value of function:

        Neither args nor return values are encoded/decoded
        '''
        conn = redis.Redis(connection_pool=self.pool)
        command = args[0]
        key = self._namespace(args[1])
        args = args[2:]
        func = getattr(conn, command)
        return func(key, *args)


    def increment(self, dict_name, key, value=1.0):
        '''
        increment the value stored at dict_name(key) by value
        '''
        conn = redis.Redis(connection_pool=self.pool)
        if isinstance(value, int):
            value = float(value)
            ## while redis will allow you to go from int to float, you
            ## cannot go back to using hincrby if a hash has floats in
            ## it.
            #conn.hincrby(self._namespace(dict_name), self._encode(key), value)
        if isinstance(value, float):
            conn.hincrbyfloat(self._namespace(dict_name), self._encode(key), value)
        else:
            raise TypeError('%r is not int or float' % value)




'''Redis-based distributed work manager.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

Rejester is a distributed work manager, using the `Redis`_ in-memory
database as shared state.  In typical use, some number of systems run
rejester *worker* processes, which query the central database to find
new work; anything that can reach the database can submit jobs.  The
standard rejester work system can also directly deliver work units to
programs without using the dedicated worker process.  For specialized
applications there is also a globally atomic string-based priority
queue.

.. _Redis: http://redis.io/

Task system
===========

Core API
--------

.. autoclass:: TaskMaster
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: WorkUnit
   :members:
   :undoc-members:
   :show-inheritance:

Workers
-------

.. autoclass:: Worker
   :members:
   :undoc-members:
   :show-inheritance:

.. automodule:: rejester.workers
   :members:
   :undoc-members:
   :show-inheritance:

.. currentmodule:: rejester

Implementation Details
----------------------

.. autoclass:: Registry
   :members:
   :undoc-members:
   :show-inheritance:

Priority queue system
=====================

.. autoclass:: RejesterQueue
   :members:
   :undoc-members:
   :show-inheritance:

Exceptions
==========

.. automodule:: rejester.exceptions
   :members:
   :undoc-members:
   :show-inheritance:

'''
from __future__ import absolute_import
from rejester.config import config_name, default_config, add_arguments, \
    runtime_keys, check_config
from rejester._task_master import TaskMaster, WorkUnit, Worker
from rejester._queue import RejesterQueue
from rejester._registry import Registry



"""Standalone test for 'rejester run_worker'.

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.

"""

from __future__ import absolute_import
import errno
import logging
import os
import signal
import sys
import time

import pexpect
import pytest

logger = logging.getLogger(__name__)
pytest_plugins = 'rejester.tests.fixtures'

@pytest.fixture
def rejester_cli_namespace(request, namespace_string, _rejester_config):
    """a rejester namespace that deletes itself using 'rejester delete'"""
    redis = _rejester_config['registry_addresses'][0]
    def fin():
        pexpect.run('rejester --registry-address {} --namespace {} delete --yes'
                    .format(redis, namespace_string),
                    logfile=sys.stdout, timeout=5)
    request.addfinalizer(fin)
    return namespace_string

def test_run(tmpdir, rejester_cli_namespace, _rejester_config):
    """Lifecycle test for 'rejester run_worker'.

    Running the top-level program starts a worker as a daemon process,
    so we need to verify that (1) the 'rejester' subprocess exits,
    (2) the daemon starts up, and (3) the daemon dies when signaled.

    """
    namespace = rejester_cli_namespace
    redis = _rejester_config['registry_addresses'][0]
    tmp_pid = str(tmpdir.join('pid'))
    tmp_log = str(tmpdir.join('log'))
    logger.info('pidfile=%r', tmp_pid)
    pexpect.run('rejester --registry-address {} --namespace {} run_worker '
                '--pidfile {} --logpath {}'
                .format(redis, namespace, tmp_pid, tmp_log),
                logfile=sys.stdout,
                timeout=5)
    # This spawns a daemon and exits, so run() should return promptly,
    # and the pid file should exist
    assert os.path.exists(tmp_pid)
    pid = int(open(tmp_pid, 'r').read())
    # Unixese for "make sure this process exists"
    os.kill(pid, 0)
    try:
        logger.info('Sending SIGTERM to pid {0}'.format(pid))
        os.kill(pid, signal.SIGTERM)
        start_time = 0
        end_time = time.time() + 20
        while time.time() < end_time:
            try:
                os.kill(pid, 0)
            except OSError, exc:
                assert exc.errno == errno.ESRCH
                break
            logger.debug('{:1f} elapsed seconds, pid {} still alive'
                         .format(time.time() - start_time, pid))
            time.sleep(0.2)  # wait a moment for daemon child to exit
        assert time.time() < end_time, "Daemon process did not die when asked"
    finally:
        # kill -9 pid, if it still exists
        try:
            os.kill(pid, signal.SIGKILL)
        except OSError, exc:
            pass



'''External API for rejester task system.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2013 Diffeo, Inc.

'''
from __future__ import absolute_import
from __future__ import division
import abc
import uuid
import time
import logging
import psutil
import socket
import traceback
import pkg_resources
from operator import itemgetter

from rejester._registry import Registry
from rejester.exceptions import ProgrammerError, LockError, \
    LostLease, EnvironmentError, NoSuchWorkSpecError

logger = logging.getLogger(__name__)

BUNDLES = 'bundles'
NICE_LEVELS = 'NICE_LEVELS'
WORK_SPECS = 'WORK_SPECS'
WORK_UNITS_ = 'WORK_UNITS_'
_BLOCKED = '_BLOCKED'
_BLOCKS = '_BLOCKS'
_DEPENDS = '_DEPENDS'
_FINISHED = '_FINISHED'
_FAILED = '_FAILED'
WORKER_STATE_ = 'WORKER_STATE_'
WORKER_OBSERVED_MODE = 'WORKER_OBSERVED_MODE'
ACTIVE_LEASES = 'ACTIVE_LEASES'

class Worker(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self, config):
        self.config = config
        self.task_master = TaskMaster(config)
        self.worker_id = None
        self.lifetime = 300 ## five minutes

    def environment(self):
        '''
        raw data about worker to support forensics on failed tasks
        '''
        env = dict(
            worker_id = self.worker_id,
            host = socket.gethostbyaddr(socket.gethostname()),
            fqdn = socket.getfqdn(),
            version = pkg_resources.get_distribution("rejester").version, # pylint: disable=E1103
            working_set = [(dist.key, dist.version) for dist in pkg_resources.WorkingSet()], # pylint: disable=E1103
            #config_hash = self.config['config_hash'],
            #config_json = self.config['config_json'],
            memory = psutil.phymem_usage(),
        )
        return env

    def register(self):
        '''record the availability of this worker and get a unique identifer
        '''
        if self.worker_id:
            raise ProgrammerError('Worker.register cannot be called again without first calling unregister; it is not idempotent')
        self.worker_id = uuid.uuid4().hex
        self.heartbeat()
        return self.worker_id

    def unregister(self):
        '''remove this worker from the list of available workers
        ''' 
        with self.task_master.registry.lock() as session:
            session.delete(WORKER_STATE_ + self.worker_id)
            session.popmany(WORKER_OBSERVED_MODE, self.worker_id)
        self.worker_id = None

    def heartbeat(self):
        '''record the current worker state in the registry and also the
        observed mode and return it

        :returns mode:
        ''' 
        mode = self.task_master.get_mode()
        with self.task_master.registry.lock() as session:
            session.set(WORKER_OBSERVED_MODE, self.worker_id, mode,
                        priority=time.time() + self.lifetime)
            session.update(WORKER_STATE_ + self.worker_id, self.environment(), 
                           expire=self.lifetime)
        logger.debug('worker {} observed mode {}'.format(self.worker_id, mode))
        return mode

    @abc.abstractmethod
    def run(self):
        return


class WorkUnit(object):
    def __init__(self, registry, work_spec_name, key, data, worker_id=None, expires=None):
        if not worker_id:
            raise ProgrammerError('must specify a worker_id, not: %r' % worker_id)
        self.worker_id = worker_id
        self.registry = registry
        self.work_spec_name = work_spec_name
        self.key = key
        self.data = data
        self.finished = False
        self.failed = False
        self.expires = expires  # time.time() when lease expires
        self._spec_cache = None  # storage for lazy getter property
        self._module_cache = None  # storage for lazy getter property
        logger.debug('WorkUnit init %r', self.key)

    def __repr__(self):
        return 'WorkUnit(key=%s)' % self.key

    # lazy caching getter for work spec
    @property
    def spec(self):
        if self._spec_cache is None:
            self._spec_cache = self.registry.get(WORK_SPECS, self.work_spec_name)
        return self._spec_cache

    # lazy caching getter for module
    @property
    def module(self):
        if self._module_cache is None:
            funclist = filter(None, (self.spec.get('run_function'), self.spec.get('terminate_function')))
            if funclist:
                try:
                    self._module_cache = __import__(
                        self.spec['module'], globals(), (), funclist, -1)
                except Exception, exc:
                    logger.error('failed to load spec["module"] = %r',
                                 self.spec['module'], exc_info=True)
                    raise
        return self._module_cache

    def run(self):
        '''execute this WorkUnit using the function specified in its
        work_spec.  Is called multiple times.
        '''
        run_function = getattr(self.module, self.spec['run_function'])
        logger.info('running work unit {}'.format(self.key))
        try:
            ret_val = run_function(self)
            self.update()
            logger.info('completed work unit {}'.format(self.key))
            return ret_val
        except Exception, exc:
            logger.error('work unit {} failed'.format(self.key),
                         exc_info=True)
            raise

    def terminate(self):
        '''shutdown this WorkUnit using the function specified in its
        work_spec.  Is called multiple times.
        '''
        terminate_function_name = self.spec.get('terminate_function')
        if not terminate_function_name:
            logger.error('tried to terminate WorkUnit(%r) but no function name', self.key)
            return None
        terminate_function = getattr(self.module, self.spec['terminate_function'])
        if not terminate_function:
            logger.error('tried to terminate WorkUnit(%r) but no function %s in module %r', self.key, terminate_function_name, self.module.__name__)
            return None
        logger.info('calling terminate function for work unit {}'
                    .format(self.key))
        ret_val = terminate_function(self)
        self.update(lease_time=-10)
        return ret_val

    def update(self, lease_time=300):
        '''reset the task lease by incrementing lease_time.  Default is to put
        it five minutes ahead of now.  Setting lease_time<0 will drop
        the lease.

        If the worker waited too long since the previous call to
        update and another worker has taken this WorkUnit, then this
        raises rejester.exceptions.LostLease
        '''
        if self.finished:
            raise ProgrammerError('cannot .update() after .finish()')
        if self.failed:
            raise ProgrammerError('cannot .update() after .fail()')
        with self.registry.lock() as session:
            if self.finished:
                logger.debug('WorkUnit(%r) became finished while waiting for lock to update', self.key)
                return
            if self.failed:
                logger.debug('WorkUnit(%r) became failed while waiting for lock to update', self.key)
                return
            try:
                self.expires = time.time() + lease_time
                session.update(
                    WORK_UNITS_ + self.work_spec_name,
                    {self.key: self.data}, 
                    priorities={self.key: self.expires},
                    locks={self.key: self.worker_id})

            except EnvironmentError, exc:
                raise LostLease(exc)
                
    def finish(self):
        '''move this WorkUnit to finished state
        '''
        if self.finished or self.failed:
            return
        with self.registry.lock() as session:
            if self.finished or self.failed:
                return
            logger.debug('WorkUnit finish %r', self.key)
            session.move(
                WORK_UNITS_ + self.work_spec_name,
                WORK_UNITS_ + self.work_spec_name + _FINISHED,
                {self.key: self.data})
            blocks = session.get(WORK_UNITS_ + self.work_spec_name + _BLOCKS,
                                 self.key)
            if blocks is not None:
                for block in blocks:
                    spec = block[0]
                    unit = block[1]
                    hard = block[2]
                    depends = session.get(WORK_UNITS_ + spec + _DEPENDS, unit)
                    if depends is None: continue
                    depends.remove([self.work_spec_name, self.key])
                    if len(depends) == 0:
                        session.popmany(WORK_UNITS_ + spec + _DEPENDS, unit)
                        unitdef = session.get(WORK_UNITS_ + spec + _BLOCKED,
                                              unit)
                        session.move(WORK_UNITS_ + spec + _BLOCKED,
                                     WORK_UNITS_ + spec,
                                     { unit: unitdef })
                    else:
                        session.set(WORK_UNITS_ + spec + _DEPENDS, unit,
                                    depends)
        self.finished = True

    def fail(self, exc=None):
        '''move this WorkUnit to failed state and record info about the
        worker
        '''
        if self.failed or self.finished:
            return
        self.data['traceback'] = exc and traceback.format_exc(exc) or exc
        #self.data['worker_state'] = self.worker_state
        with self.registry.lock() as session:
            if self.finished or self.failed:
                return
            session.move(
                WORK_UNITS_ + self.work_spec_name,
                WORK_UNITS_ + self.work_spec_name + _FAILED,
                {self.key: self.data})
            blocks = session.get(WORK_UNITS_ + self.work_spec_name + _BLOCKS,
                                 self.key)
            if blocks is not None:
                for block in blocks:
                    spec = block[0]
                    unit = block[1]
                    hard = block[2]
                    if hard:
                        session.popmany(WORK_UNITS_ + spec + _DEPENDS, unit)
                        unitdef = session.get(WORK_UNITS_ + spec + _BLOCKED,
                                              unit)
                        if unitdef is not None:
                            session.move(WORK_UNITS_ + spec + _BLOCKED,
                                         WORK_UNITS_ + spec + _FAILED,
                                         { unit: unitdef })
                    else:
                        depends = session.get(WORK_UNITS_ + spec + _DEPENDS, unit)
                        if depends is None: continue
                        depends.remove([self.work_spec_name, self.key])
                        if len(depends) == 0:
                            session.popmany(WORK_UNITS_ + spec + _DEPENDS, unit)
                            unitdef = session.get(WORK_UNITS_ + spec + _BLOCKED,
                                                  unit)
                            if unitdef is not None:
                                session.move(WORK_UNITS_ + spec + _BLOCKED,
                                             WORK_UNITS_ + spec,
                                             { unit: unitdef })
                        else:
                            session.set(WORK_UNITS_ + spec + _DEPENDS, unit,
                                        depends)
        self.failed = True


class TaskMaster(object):
    '''Control object for the rejester task queue.

    The task queue consists of a series of *work specs*, which include
    configuration information, and each work spec has some number of
    *work units* attached to it.  Both the work specs and work units
    are defined as (non-empty) dictionaries.  The work spec must have
    keys 'name' and 'min_gb'; in typical use, 'module' names a Python
    module, 'run_function' and 'terminate_function' name specific
    functions within that module, and 'config' provides a
    configuration dictionary as well.

    A work unit can be in one of five states.  It is *available* if it
    has been added to the queue but nobody is working on it.  It is
    *pending* if somebody is currently working on it.  When they
    finish, it will become either *finished* or *failed*.  If
    dependencies are added between tasks using `add_dependent_task()`,
    a task can also be *blocked*.

    The general flow for a rejester application is to
    `set_mode(IDLE)`, then add work units using `update_bundle()`, and
    `set_mode(RUN)`.  `get_work()` will return work units until all
    have been consumed.  `set_mode(TERMINATE)` will instruct workers
    to shut down.

    This object keeps very little state locally and can safely be used
    concurrently, including from multiple systems.  Correspondingly,
    any settings here, including `set_mode()`, are persistent even
    beyond the end of the current process.

    '''

    def __init__(self, config):
        config['app_name'] = 'rejester'
        self.config = config
        self.registry = Registry(config)

    RUN = 'RUN'
    IDLE = 'IDLE'
    TERMINATE = 'TERMINATE'

    def set_mode(self, mode):
        '''Set the global mode of the rejester system.

        This must be one of the constants `TERMINATE`, `RUN`, or `IDLE`.
        `TERMINATE` instructs any running workers to do an orderly
        shutdown.  `IDLE` instructs workers to stay running but not
        start new jobs.  `RUN` tells workers to do actual work.

        :param str mode: new rejester mode
        :raise ProgrammerError: on invalid `mode`

        '''
        if not mode in [self.TERMINATE, self.RUN, self.IDLE]:
            raise ProgrammerError('mode=%r is not recognized' % mode)
        with self.registry.lock() as session:
            session.set('modes', 'mode', mode)
        logger.info('set mode to %s', mode)

    def get_mode(self):
        '''Get the global mode of the rejester system.

        :return: rejester mode, `IDLE` if unset

        '''
        with self.registry.lock() as session:
            return session.get('modes', 'mode') or self.IDLE

    def idle_all_workers(self):
        self.set_mode(self.IDLE)
        while 1:
            num_pending = dict()
            for work_spec_name in self.registry.pull(NICE_LEVELS).keys():
                num_pending[work_spec_name] = self.num_pending(work_spec_name)
            if sum(num_pending.values()) == 0:
                break
            logger.warn('waiting for pending work_units: %r', num_pending)
            time.sleep(1)

    def mode_counts(self):
        '''counts the modes observed by workers who heartbeated within one
        lifetime of the present time.
        '''
        modes = {self.RUN: 0, self.IDLE: 0, self.TERMINATE: 0}
        for worker_id, mode in self.workers().items():
            modes[mode] += 1
        return modes

    def workers(self, alive=True):
        '''returns dictionary of all worker_ids and current mode.  If
        alive=True, then only include worker_ids that have heartbeated
        within one lifetime of now.

        '''
        with self.registry.lock() as session:
            return session.filter(
                WORKER_OBSERVED_MODE, 
                priority_min=alive and time.time() or None)

    def dump(self):
        '''Print the entire contents of this to debug log messages.

        This is really only intended for debugging.  It could produce
        a lot of data.

        '''
        with self.registry.lock() as session:
            for work_spec_name in self.registry.pull(NICE_LEVELS).iterkeys():
                def scan(sfx):
                    v = self.registry.pull(WORK_UNITS_ + work_spec_name + sfx)
                    if v is None: return []
                    return v.keys()
                for key in scan(''):
                    logger.debug('spec {} unit {} available or pending'
                                 .format(work_spec_name, key))
                for key in scan(_BLOCKED):
                    blocked_on = self.registry.get(
                        WORK_UNITS_ + work_spec_name + _DEPENDS, key)
                    logger.debug('spec {} unit {} blocked on {!r}'
                                 .format(work_spec_name, key, blocked_on))
                for key in scan(_FINISHED):
                    logger.debug('spec {} unit {} finished'
                                 .format(work_spec_name, key))
                for key in scan(_FAILED):
                    logger.debug('spec {} unit {} failed'
                                 .format(work_spec_name, key))


    @classmethod
    def validate_work_spec(cls, work_spec):
        'raise ProgrammerError if work_spec is invalid'
        if 'name' not in work_spec:
            raise ProgrammerError('work_spec lacks "name"')
        if 'min_gb' not in work_spec or \
                not isinstance(work_spec['min_gb'], (float, int)):
            raise ProgrammerError('work_spec["min_gb"] must be a number')

    def num_available(self, work_spec_name):
        return self.registry.len(WORK_UNITS_ + work_spec_name, 
                                 priority_max=time.time())

    def num_pending(self, work_spec_name):
        return self.registry.len(WORK_UNITS_ + work_spec_name, 
                                 priority_min=time.time())

    def num_blocked(self, work_spec_name):
        return self.registry.len(WORK_UNITS_ + work_spec_name + _BLOCKED)

    def num_finished(self, work_spec_name):
        return self.registry.len(WORK_UNITS_ + work_spec_name + _FINISHED)

    def num_failed(self, work_spec_name):
        return self.registry.len(WORK_UNITS_ + work_spec_name + _FAILED)

    def num_tasks(self, work_spec_name):
        return self.num_finished(work_spec_name) + \
               self.num_failed(work_spec_name) + \
               self.registry.len(WORK_UNITS_ + work_spec_name)

    def status(self, work_spec_name):
        return dict(
            num_available=self.num_available(work_spec_name),
            num_pending=self.num_pending(work_spec_name),
            num_blocked=self.num_blocked(work_spec_name),
            num_finished=self.num_finished(work_spec_name),
            num_failed=self.num_failed(work_spec_name),
            num_tasks=self.num_tasks(work_spec_name),
            )


    def list_work_units(self, work_spec_name):
        """Get a dictionary of work units for some work spec.

        The dictionary is from work unit name to work unit definiton.
        Only work units that have not been completed ("available" or
        "pending" work units) are included.

        """
        with self.registry.lock(atime=1000) as session:
            return session.pull(WORK_UNITS_ + work_spec_name)

    def inspect_work_unit(self, work_spec_name, work_unit_key):
        '''Get the data for some work unit.

        Returns the data for that work unit, or `None` if it really
        can't be found.

        :param str work_spec_name: name of the work spec
        :param str work_unit_key: name of the work unit
        :return: definition of the work unit, or `None`
        '''
        with self.registry.lock(atime=1000) as session:
            work_unit_data = session.get(
                WORK_UNITS_ + work_spec_name, work_unit_key)
            if not work_unit_data:
                work_unit_data = session.get(
                    WORK_UNITS_ + work_spec_name + _BLOCKED, work_unit_key)
            if not work_unit_data:
                work_unit_data = session.get(
                    WORK_UNITS_ + work_spec_name + _FINISHED, work_unit_key)
            if not work_unit_data:
                work_unit_data = session.get(
                    WORK_UNITS_ + work_spec_name + _FAILED, work_unit_key)
            return work_unit_data

    def _inspect_work_unit(self, session, work_spec_name, work_unit_key):
        return work_unit_data


    def reset_all(self, work_spec_name):
        self.idle_all_workers()
        with self.registry.lock(ltime=10000) as session:
            session.move_all(WORK_UNITS_ + work_spec_name + _FINISHED,
                             WORK_UNITS_ + work_spec_name)
            session.reset_priorities(WORK_UNITS_ + work_spec_name, 0)

    def update_bundle(self, work_spec, work_units, nice=0):
        '''Load a work spec and some work units into the task list.
        
        update the work_spec and work_units.  Overwrites any existing
        work_spec with the same ``work_spec['name']`` and similarly
        overwrites any WorkUnit with the same ``work_unit.key``

        :param work_units: dict of dicts where the keys are used as
          WorkUnit.key and the values are used as WorkUnit.data

        '''
        self.validate_work_spec(work_spec)
        
        work_spec_name = work_spec['name']
        with self.registry.lock(atime=1000) as session:
            session.update(NICE_LEVELS, {work_spec_name: nice})
            session.update(WORK_SPECS,  {work_spec_name: work_spec})
            
            ## redis chokes on lua scripts with 
            work_units = work_units.items()
            window_size = 10**4
            for i in xrange(0, len(work_units), window_size):
                self.registry.re_acquire_lock(ltime=2000)
                session.update(WORK_UNITS_ + work_spec_name, 
                               dict(work_units[i: i + window_size]))

    def add_dependent_work_units(self, work_unit, depends_on, hard=True):
        """Add work units, where one prevents execution of the other.

        The two work units may be attached to different work specs,
        but both must be in this task master's namespace.  `work_unit`
        and `depends_on` are both tuples of (work spec name, work unit
        name, work unit dictionary).  The work specs must already
        exist; they may be created with :meth:`update_bundle` with
        an empty work unit dictionary.  If a work unit dictionary is
        provided with either work unit, then this defines that work
        unit, and any existing definition is replaced.  Either or both
        work unit dictionaries may be :const:`None`, in which case the
        work unit is not created if it does not already exist.  In
        this last case, the other work unit will be added if
        specified, but the dependency will not be added, and this
        function will return :const:`False`.  In all other cases, this
        dependency is added in addition to all existing dependencies
        on either or both work units, even if the work unit dictionary
        is replaced.

        `work_unit` will not be executed or reported as available via
        :meth:`get_work` until `depends_on` finishes execution.  If
        the `depends_on` task fails, then the `hard` parameter
        describes what happens: if `hard` is :const:`True` then
        `work_unit` will also fail, but if `hard` is :const:`False`
        then `work_unit` will be able to execute even if `depends_on`
        fails, it just must have completed some execution attempt.

        Calling this function with ``hard=True`` suggests an ordered
        sequence of tasks where the later task depends on the output
        of the earlier tasks.  Calling this function with
        ``hard=False`` suggests a cleanup task that must run after
        this task (and, likely, several others) are done, but doesn't
        specifically depend on its result being available.

        :param work_unit: "Later" work unit to execute
        :paramtype work_unit: tuple of (str,str,dict)
        :param depends_on: "Earlier" work unit to execute
        :paramtype depends_on: tuple of (str,str,dict)
        :param bool hard: if True, then `work_unit` automatically fails
          if `depends_on` fails
        :return: :const:`True`, unless one or both of the work units
          didn't exist and weren't specified, in which case, :const:`False`
        :raise NoSuchWorkSpecError: if a work spec was named that doesn't
          exist

        """
        # There's no good, not-confusing terminology here.
        # I'll call work_unit "later" and depends_on "earlier"
        # consistently, because that at least makes the time flow
        # correct.
        later_spec, later_unit, later_unitdef = work_unit
        earlier_spec, earlier_unit, earlier_unitdef = depends_on
        with self.registry.lock(atime=1000) as session:
            # Bail if either work spec doesn't already exist
            if session.get(WORK_SPECS, later_spec) is None:
                raise NoSuchWorkSpecError(later_spec)
            if session.get(WORK_SPECS, earlier_spec) is None:
                raise NoSuchWorkSpecError(earlier_spec)
            
            # Cause both work units to exist (if possible)
            # Note that if "earlier" is already finished, we may be
            # able to make "later" available immediately
            earlier_done = False
            earlier_successful = False
            if earlier_unitdef is not None:
                session.update(WORK_UNITS_ + earlier_spec,
                               { earlier_unit: earlier_unitdef })
            else:
                earlier_unitdef = session.get(
                    WORK_UNITS_ + earlier_spec, earlier_unit)
                if earlier_unitdef is None:
                    earlier_unitdef = session.get(
                        WORK_UNITS_ + earlier_spec + _BLOCKED, earlier_unit)
                if earlier_unitdef is None:
                    earlier_unitdef = session.get(
                        WORK_UNITS_ + earlier_spec + _FINISHED, earlier_unit)
                    if earlier_unitdef is not None:
                        earlier_done = True
                        earlier_successful = True
                if earlier_unitdef is None:
                    earlier_unitdef = session.get(
                        WORK_UNITS_ + earlier_spec + _FAILED, earlier_unit)
                    if earlier_unitdef is not None:
                        earlier_done = True

            later_failed = earlier_done and hard and not earlier_successful
            later_unblocked = ((earlier_done and not later_failed) or
                               (earlier_unitdef is None))
            if later_failed:
                later_destination = WORK_UNITS_ + later_spec + _FAILED
            elif later_unblocked:
                later_destination = WORK_UNITS_ + later_spec
            else:
                later_destination = WORK_UNITS_ + later_spec + _BLOCKED

            if later_unitdef is not None:
                for suffix in ['', _FINISHED, _FAILED, _BLOCKED]:
                    k = WORK_UNITS_ + later_spec + suffix
                    if k != later_destination:
                        session.popmany(k, later_unit)
                session.update(later_destination,
                               { later_unit: later_unitdef })
            elif earlier_unitdef is not None:
                later_unitdef = session.get(
                    WORK_UNITS_ + later_spec, later_unit)
                if later_unitdef is not None:
                    session.move(
                        WORK_UNITS_ + later_spec,
                        WORK_UNITS_ + later_spec + _BLOCKED,
                        { later_unit: later_unitdef })
                else:
                    later_unitdef = session.get(
                        WORK_UNITS_ + later_spec + _BLOCKED, later_unit)

            if later_unitdef is None or earlier_unitdef is None:
                return False

            # Now both units exist and are in the right place;
            # record the dependency
            blocks = session.get(WORK_UNITS_ + earlier_spec + _BLOCKS,
                                 earlier_unit)
            if blocks is None: blocks = []
            blocks.append([later_spec, later_unit, hard])
            session.set(WORK_UNITS_ + earlier_spec + _BLOCKS,
                        earlier_unit, blocks)

            depends = session.get(WORK_UNITS_ + later_spec + _DEPENDS,
                                  later_unit)
            if depends is None: depends = []
            depends.append([earlier_spec, earlier_unit])
            session.set(WORK_UNITS_ + later_spec + _DEPENDS,
                        later_unit, depends)

            return True

    def nice(self, work_spec_name, nice):        
        with self.registry.lock(atime=1000) as session:
            session.update(NICE_LEVELS, dict(work_spec_name=nice))

    def get_work(self, worker_id, available_gb=None, lease_time=300):
        '''obtain a WorkUnit instance based on available memory for the
        worker process.  
        
        :param worker_id: unique identifier string for a worker to
          which a WorkUnit will be assigned, if available.
        :param available_gb: number of gigabytes of RAM available to
          this worker
        :param lease_time: how many seconds to lease a WorkUnit

        '''
        start = time.time()

        if not isinstance(available_gb, (int, float)):
            raise ProgrammerError('must specify available_gb')

        work_unit = None

        try: 
            with self.registry.lock(atime=1000, ltime=10000) as session:
                ## figure out which work_specs have low nice levels
                nice_levels = session.pull(NICE_LEVELS)
                nice_levels = nice_levels.items()
                nice_levels.sort(key=itemgetter(1))

                work_specs = session.pull(WORK_SPECS)
                for work_spec_name, nice in nice_levels:
                    self.registry.re_acquire_lock(ltime=10000)

                    ## verify sufficient memory
                    if available_gb < work_specs[work_spec_name]['min_gb']:
                        continue

                    logger.debug('considering %s %s', work_spec_name, nice_levels)

                    ## try to get a task
                    wu_expires = time.time() + lease_time
                    _work_unit = session.getitem_reset(
                        WORK_UNITS_ + work_spec_name,
                        priority_max=time.time(),
                        new_priority=wu_expires,
                        lock=worker_id,
                    )

                    if _work_unit:
                        logger.info('work unit %r', _work_unit)
                        work_unit = WorkUnit(
                            self.registry, work_spec_name,
                            _work_unit[0], _work_unit[1],
                            worker_id=worker_id,
                            expires=wu_expires,
                        )
                        break

        except (LockError, EnvironmentError), exc:
            logger.critical('failed to get work', exc_info=True)

        logger.debug('get_work %s in %.3f', work_unit is not None, time.time() - start)
        return work_unit

    def get_assigned_work_unit(
            self, worker_id, work_spec_name, work_unit_key,
        ):
        '''get a specific WorkUnit that has already been assigned to a
        particular worker_id
        '''
        with self.registry.lock(atime=1000, ltime=10000) as session:
            assigned_work_unit_key = session.get(
                WORK_UNITS_ + work_spec_name + '_locks', worker_id)
            if not assigned_work_unit_key == work_unit_key:
                ## raise LostLease instead of EnvironmentError, so
                ## users of TaskMaster can have a single type of
                ## expected exception, rather than two
                raise LostLease(
                    'assigned_work_unit_key=%r != %r'
                    % (assigned_work_unit_key, work_unit_key))
            ## could trap EnvironmentError and raise LostLease instead
            work_unit_data = session.get(
                        WORK_UNITS_ + work_spec_name,
                        work_unit_key)
            return WorkUnit(
                self.registry, work_spec_name,
                work_unit_key, work_unit_data,
                worker_id=worker_id,                            
            )



'''Tests for the 'rejester' command-line tool.

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import errno
import logging
import json
import os
import re
import signal
import sys
import time

import pexpect
import pytest
import yaml

import rejester

logger = logging.getLogger(__name__)
pytest_plugins = 'rejester.tests.fixtures'

## 1KB sized work_spec config
work_spec = dict(
    name = 'tbundle',
    desc = 'a test work bundle',
    min_gb = 8,
    config = dict(many=' ', params=''),
    module = 'rejester.workers',
    run_function = 'test_work_program',
    terminate_function = 'test_work_program',
)

def test_cli(task_master, tmpdir):
    tmpf = str(tmpdir.join("work_spec.yaml"))
    with open(tmpf, 'w') as f:
        f.write(yaml.dump(work_spec))
    assert yaml.load(open(tmpf, 'r')) == work_spec

    num_units = 11
    tmpf2 = str(tmpdir.join("work_units.yaml"))
    with open(tmpf2, 'w') as f:
        for x in xrange(num_units):
            work_unit = json.dumps({'key-%d' % x: dict(sleep=0.2, config=task_master.registry.config)})
            f.write(work_unit + '\n')

    namespace = task_master.registry.config['namespace']
    redis = task_master.registry.config['registry_addresses'][0]
    def r(action, params):
        return ('rejester --registry-address {} --app-name rejester_test '
                '--namespace {} {} {}'.format(redis, namespace, action, params))
    child = pexpect.spawn(r('load', '-u {} -w {}'.format(tmpf2, tmpf)),
                          logfile=sys.stdout)
    try:
        child.expect('loading', timeout=5)
        child.expect('pushing', timeout=5)
        child.expect('finish', timeout=5)
    finally:
        child.close()

    logger.debug(json.dumps(task_master.status(work_spec['name']), indent=4))

    out = pexpect.run(r('status', '-w {}'.format(tmpf)),
                      timeout=5, logfile=sys.stdout)
    assert re.search('num_available.*{0}'.format(num_units), out)

    out = pexpect.run(r('work_units', '-w {}'.format(tmpf)),
                      timeout=5, logfile=sys.stdout)
    assert out == ''.join(sorted(['key-{}\r\n'.format(n)
                                  for n in xrange(num_units)]))

    tmp_pid = str(tmpdir.join('pid'))
    tmp_log = str(tmpdir.join('log'))
    out = pexpect.run(r('run_worker',
                        '--logpath {} --pidfile {}'.format(tmp_log, tmp_pid)),
                      timeout=5, logfile=sys.stdout)
    assert os.path.exists(tmp_pid)
    pid = int(open(tmp_pid).read())
    try:
        os.kill(pid, 0) # will raise OSError if pid is dead

        out = pexpect.run(r('set_RUN', ''),
                          timeout=5, logfile=sys.stdout) 
        assert out.find('set') != -1

        start_time = time.time()
        end_time = start_time + 60
        while time.time() < end_time:
            if task_master.num_finished(work_spec['name']) == num_units:
                break
            if os.path.exists(tmp_log):
                with open(tmp_log, 'r') as f:
                    print f.read()
            logger.info('{:.1f} elapsed seconds, {} finished'
                        .format(time.time() - start_time,
                                task_master.num_finished(work_spec['name'])))
            logger.debug(json.dumps(task_master.status(work_spec['name'])))
            time.sleep(1)

        assert task_master.num_finished(work_spec['name']) == num_units
        logger.info('tasks completed')
    finally:
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError, exc:
            pass # assume errno == -ESRCH; we'll do more below
        start_time = time.time()
        end_time = start_time + 20
        while time.time() < end_time:
            try:
                os.kill(pid, 0)
            except OSError, exc:
                # If we get a "no such process" error then the process is dead
                # (which is what we want)
                assert exc.errno == errno.ESRCH
                break
            logger.debug('{:.1f} elapsed seconds, {} still alive'
                         .format(time.time() - start_time, pid))
            time.sleep(0.2)  # wait a moment for daemon child to exit
        # Double-check the process is dead
        try:
            os.kill(pid, 0)
            # if we did not get an exception then the process is alive
            logger.warn('worker pid {0} did not respond to SIGTERM, killing'
                        .format(pid))
            os.kill(pid, signal.SIGKILL)
            assert False, "workers did not shut down"
        except OSError, exc:
            assert exc.errno == errno.ESRCH
            pass
        logger.info('worker exited')



'''
This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import contextlib
import os
import copy
import time
import logging
import multiprocessing
import sys

import pytest

import rejester
from rejester.workers import run_worker, MultiWorker

logger = logging.getLogger(__name__)
pytest_plugins = 'rejester.tests.fixtures'

def test_task_register(task_master):
    worker = MultiWorker(task_master.registry.config)
    worker_id = worker.register()
    assert worker_id in task_master.workers()
    worker.unregister()
    assert worker_id not in task_master.workers()


def work_program(work_unit):
    ## just to show that this works, we get the config from the data
    ## and *reconnect* to the registry with a second instances instead
    ## of using work_unit.registry
    config = work_unit.data['config']
    sleeptime = float(work_unit.data.get('sleep', 9.0))
    task_master = rejester.TaskMaster(config)
    logger.info('executing work_unit %r ... %s', work_unit.key, sleeptime)
    time.sleep(sleeptime)  # pretend to work

def work_program_broken(work_unit):
    logger.info('executing "broken" work_unit %r', work_unit.key)

    ## just to show that this works, we get the config from the data
    ## and *reconnect* to the registry with a second instances instead
    ## of using work_unit.registry
    config = work_unit.data['config']
    task_master = rejester.TaskMaster(config)
    raise Exception('simulate broken work_unit')

## 1KB sized work_spec config
work_spec = dict(
    name = 'tbundle',
    desc = 'a test work bundle',
    min_gb = 8,
    config = dict(many=' ' * 2**10, params=''),
    module = 'rejester.tests.test_workers',
    run_function = 'work_program',
    terminate_function = 'work_program',
)

@contextlib.contextmanager
def run_multi_worker(task_master, duration):
    """Spawn a MultiWorker and clean up from it sanely.

    Use this as:

    >>> with run_multi_worker(task_master, 60) as deadline:
    ...   while time.time() < deadline:
    ...     # do some work
    ...     time.sleep(1)

    On exiting the context block, this will do a very clean shutdown
    of the workers; if anything goes wrong this will do an unclean shutdown
    of the workers.

    """
    p = multiprocessing.Process(target=run_worker, 
                                args=(MultiWorker, task_master.registry.config))
    num_workers = multiprocessing.cpu_count()
    logger.debug('expecting num_workers=%d', num_workers)

    start = time.time()
    end = start + duration
    finished_cleanly = False
    already_set_idle = False
    already_set_terminate = False
    p.start()

    try:
        yield end

        while time.time() < end:
            modes = task_master.mode_counts()
            logger.debug(modes)

            if modes[task_master.RUN] >= num_workers:
                logger.info('setting mode to IDLE')
                task_master.idle_all_workers()
                already_set_idle = True

            if modes[task_master.IDLE] >= 0 and already_set_idle:
                logger.info('setting mode to TERMINATE')
                task_master.set_mode(task_master.TERMINATE)
                already_set_terminate = True
        
            if p.exitcode is not None:
                assert already_set_idle
                assert already_set_terminate
                if p.exitcode == 0:
                    finished_cleanly = True
                break

            time.sleep(2)

        assert finished_cleanly, ('timed out after {0} seconds'
                                  .format(time.time() - start))

        p.join()
        logger.info('finished running %d worker processes', num_workers)
    finally:
        if p.is_alive():
            logger.debug("killing worker processes pid={0}".format(p.pid))
            p.terminate()
            p.join(1.0)
        if p.is_alive():
            logger.warn("worker processes pid={0} failed to die, hard killing"
                        .format(p.pid))
            os.kill(p.pid, signal.SIGKILL)
            p.join(1.0)
        if p.is_alive():
            logger.critical("worker processes pid={0} resisted SIGKILL"
                            .format(p.pid))

def test_task_master_multi_worker(task_master):
    num_units = 10
    num_units_cursor = 0
    work_units = {'key' + str(x): dict(config=task_master.registry.config) 
                  for x in xrange(num_units_cursor, num_units_cursor + num_units)}
    num_units_cursor += num_units
    task_master.update_bundle(work_spec, work_units)

    task_master.set_mode(task_master.RUN)

    with run_multi_worker(task_master, 60):
        pass

    num_workers = multiprocessing.cpu_count()
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) >= num_workers


def test_task_master_multi_worker_failed_task(task_master):
    work_spec_broken = copy.deepcopy(work_spec)
    work_spec_broken['run_function'] = 'work_program_broken'

    num_units = 10
    num_units_cursor = 0
    work_units = {'key' + str(x): dict(config=task_master.registry.config) 
                  for x in xrange(num_units_cursor, num_units_cursor + num_units)}
    num_units_cursor += num_units
    task_master.update_bundle(work_spec_broken, work_units)

    task_master.set_mode(task_master.RUN)

    with run_multi_worker(task_master, 60):
        pass

    num_workers = multiprocessing.cpu_count()
    assert task_master.num_failed(work_spec['name']) >= num_workers
    assert task_master.num_finished(work_spec['name']) == 0


@pytest.mark.slow
def test_task_master_multi_worker_multi_update(task_master):
    num_units = 10
    num_units_cursor = 0
    workduration = 2.0
    work_units = {'key' + str(x): dict(config=task_master.registry.config, sleep=workduration)
                  for x in xrange(num_units_cursor, num_units_cursor + num_units)}
    num_units_cursor += num_units
    task_master.update_bundle(work_spec, work_units)

    task_master.set_mode(task_master.RUN)

    with run_multi_worker(task_master, 120) as end:

        logger.info('waiting for first set to finish')
        while time.time() < end:
            assert task_master.num_failed(work_spec['name']) == 0
            if task_master.num_finished(work_spec['name']) >= len(work_units):
                break

            time.sleep(2)

        # add more work
        work_units2 = {'key' + str(x): dict(config=task_master.registry.config)
                       for x in xrange(num_units_cursor, num_units_cursor + num_units)}
        task_master.update_bundle(work_spec, work_units2)

        logger.info('waiting for second set to finish')
        while time.time() < end:
            assert task_master.num_failed(work_spec['name']) == 0
            if task_master.num_finished(work_spec['name']) >= (num_units * 2):
                break

            time.sleep(2)

        logger.info('all jobs finished')

    num_workers = multiprocessing.cpu_count()
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) >= num_workers


def test_task_master_multi_worker_multi_update_miniwait(task_master):
    num_units = 10
    num_units_cursor = 0
    workduration = 2.0
    work_units = {'key' + str(x): dict(config=task_master.registry.config, sleep=workduration)
                  for x in xrange(num_units_cursor, num_units_cursor + num_units)}
    num_units_cursor += num_units
    task_master.update_bundle(work_spec, work_units)

    task_master.set_mode(task_master.RUN)

    with run_multi_worker(task_master, 120) as end:
        # just wait a little bit, presumably some set of things are running
        time.sleep(1)

        # add more work
        work_units2 = {'key' + str(x): dict(config=task_master.registry.config, sleep=workduration)
                       for x in xrange(num_units_cursor, num_units_cursor + num_units)}
        task_master.update_bundle(work_spec, work_units2)

        while time.time() < end:
            assert task_master.num_failed(work_spec['name']) == 0
            num_finished = task_master.num_finished(work_spec['name'])
            if num_finished >= (num_units * 2):
                break
            logger.debug('work spec %r finished=%s', work_spec['name'], num_finished)
            time.sleep(2)

    num_workers = multiprocessing.cpu_count()
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) >= num_workers



'''Exceptions raised in various places in rejester.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''

class RejesterException(Exception):
    'base exception for rejester package'
    pass

class EnvironmentError(RejesterException):
    '''indicates that the registry lost a lock or experienced a similar
failure that probably indicates a network or remote server failure'''
    pass

class LockError(RejesterException):
    'attempt to get a lock exceeded acquire time (atime)'
    pass

class NoSuchWorkSpecError(RejesterException):
    """A `TaskMaster` function was called with a nonexistent work spec"""
    pass

class ProgrammerError(RejesterException):
    pass

class PriorityRangeEmpty(RejesterException):
    '''
    given the priority_min/max, no item is available to be returned
    '''
    pass

class LostLease(RejesterException):
    '''worker waited too long between calls to update and another worker
got the WorkItem'''
    pass

class ItemInUseError(RejesterException):
    '''tried to add an item to a queue that was already in use'''
    pass



"""Various dependency tests for TaskMaster.

-----

This software is released under an MIT/X11 open source license.

Copyright 2014 Diffeo, Inc.

"""
from __future__ import absolute_import
import logging

import pytest

from rejester import TaskMaster
from rejester.exceptions import NoSuchWorkSpecError

logger = logging.getLogger(__name__)
pytest_plugins = 'rejester.tests.fixtures'

work_spec_a = { 'name': 'a', 'min_gb': 1 }
work_spec_b = { 'name': 'b', 'min_gb': 1 }
work_units_aa = { 'aa': { 'x': 'x' } }
work_units_bb = { 'bb': { 'x': 'x' } }
work_units_cc = { 'cc': { 'x': 'x' } }

def get_and_check_unit(task_master, spec, unit, fail=False):
    u = task_master.get_work('worker', available_gb=4)
    assert u.work_spec_name == spec
    assert u.key == unit
    if fail:
        u.fail()
    else:
        u.finish()
    return u

def get_and_check_done(task_master):
    unit = task_master.get_work('worker', available_gb=4)
    assert unit is None

def test_two_no_dependency(task_master):
    """Add two tasks, with priorities set so that they run in order"""
    task_master.update_bundle(work_spec_a, work_units_aa, nice=0)
    task_master.update_bundle(work_spec_b, work_units_bb, nice=1)
    get_and_check_unit(task_master, 'a', 'aa')
    get_and_check_unit(task_master, 'b', 'bb')
    get_and_check_done(task_master)

def test_two_simple_dependency(task_master):
    """Set a dependency so the tasks run out of order"""
    task_master.update_bundle(work_spec_a, work_units_aa, nice=0)
    task_master.update_bundle(work_spec_b, work_units_bb, nice=1)
    res = task_master.add_dependent_work_units(
        ('a', 'aa', None), ('b', 'bb', None))
    assert res == True
    get_and_check_unit(task_master, 'b', 'bb')
    get_and_check_unit(task_master, 'a', 'aa')
    get_and_check_done(task_master)

def test_two_hard_dependency(task_master):
    """If the depends_on task fails, the work_unit should fail too"""
    task_master.update_bundle(work_spec_a, work_units_aa, nice=0)
    task_master.update_bundle(work_spec_b, work_units_bb, nice=1)
    res = task_master.add_dependent_work_units(
        ('a', 'aa', None), ('b', 'bb', None), hard=True)
    assert res == True
    get_and_check_unit(task_master, 'b', 'bb', fail=True)
    get_and_check_done(task_master)
    assert task_master.num_failed('a') == 1
    assert task_master.num_failed('b') == 1

def test_two_soft_dependency(task_master):
    """If the depends_on task fails, the work_unit still runs"""
    task_master.update_bundle(work_spec_a, work_units_aa, nice=0)
    task_master.update_bundle(work_spec_b, work_units_bb, nice=1)
    res = task_master.add_dependent_work_units(
        ('a', 'aa', None), ('b', 'bb', None), hard=False)
    get_and_check_unit(task_master, 'b', 'bb', fail=True)
    get_and_check_unit(task_master, 'a', 'aa')
    get_and_check_done(task_master)
    assert task_master.num_failed('a') == 0
    assert task_master.num_failed('b') == 1

def test_two_caller_defines(task_master):
    """Actually define the tasks via the add_dependent function"""
    task_master.update_bundle(work_spec_a, {}, nice=0)
    task_master.update_bundle(work_spec_b, {}, nice=1)
    res = task_master.add_dependent_work_units(
        ('a', 'aa', {'hello': 'there'}), ('b', 'bb', {'hello': 'world'}))
    assert res == True
    unit = get_and_check_unit(task_master, 'b', 'bb')
    assert unit.data['hello'] == 'world'
    unit = get_and_check_unit(task_master, 'a', 'aa')
    assert unit.data['hello'] == 'there'
    get_and_check_done(task_master)

def test_two_caller_redefines(task_master):
    """add_dependent changes the definitions"""
    task_master.update_bundle(work_spec_a, {'aa': {'hello': 'world'}}, nice=0)
    task_master.update_bundle(work_spec_b, {'bb': {'hello': 'there'}}, nice=1)
    res = task_master.add_dependent_work_units(
        ('a', 'aa', {'hello': 'kitty'}), ('b', 'bb', {'hello': 'goodbye'}))
    assert res == True
    unit = get_and_check_unit(task_master, 'b', 'bb')
    assert unit.data['hello'] == 'goodbye'
    unit = get_and_check_unit(task_master, 'a', 'aa')
    assert unit.data['hello'] == 'kitty'
    get_and_check_done(task_master)

def test_missing_units(task_master):
    """Dependencies on units that don't exist"""
    task_master.update_bundle(work_spec_a, work_units_aa)
    task_master.update_bundle(work_spec_b, work_units_bb)
    res = task_master.add_dependent_work_units(
        ('a', 'aa', None), ('a', 'bb', None))
    assert res is False
    res = task_master.add_dependent_work_units(
        ('b', 'aa', None), ('b', 'bb', None))
    assert res is False
    with pytest.raises(NoSuchWorkSpecError):
        task_master.add_dependent_work_units(
            ('a', 'aa', None), ('c', 'cc', None))
    with pytest.raises(NoSuchWorkSpecError):
        task_master.add_dependent_work_units(
            ('c', 'cc', None), ('b', 'bb', None))

def test_missing_units_definitions(task_master):
    """Dependencies on units that don't exist, with partial definitions"""
    task_master.update_bundle(work_spec_a, work_units_aa)
    task_master.update_bundle(work_spec_b, work_units_bb)
    res = task_master.add_dependent_work_units(
        ('a', 'bb', { 'x': 'x' }), ('a', 'cc', None))
    assert res is False
    # This defines a/bb, and so
    res = task_master.add_dependent_work_units(
        ('a', 'aa', None), ('a', 'bb', None))
    assert res is True
    # Same test, other direction
    res = task_master.add_dependent_work_units(
        ('b', 'aa', None), ('b', 'cc', { 'x': 'x' }))
    assert res is False
    res = task_master.add_dependent_work_units(
        ('b', 'bb', None), ('b', 'cc', None))
    assert res is True
    # Can't create otherwise valid units in missing work specs
    with pytest.raises(NoSuchWorkSpecError):
        task_master.add_dependent_work_units(
            ('a', 'aa', None), ('c', 'cc', { 'x': 'x' }))
    with pytest.raises(NoSuchWorkSpecError):
        task_master.add_dependent_work_units(
            ('c', 'cc', { 'x': 'x' }), ('b', 'bb', { 'x': 'x' }))
    with pytest.raises(NoSuchWorkSpecError):
        task_master.add_dependent_work_units(
            ('c', 'aa', { 'x': 'x' }), ('c', 'bb', { 'x': 'x' }))
 
def test_three(task_master):
    """A depends on both B and C"""
    task_master.update_bundle(work_spec_a, work_units_aa, nice=0)
    task_master.update_bundle(work_spec_b, work_units_bb, nice=1)
    task_master.update_bundle(work_spec_b, work_units_cc, nice=2)
    # so natural order is a/aa, b/bb, b/cc
    res = task_master.add_dependent_work_units(('a', 'aa', None),
                                               ('b', 'bb', None))
    assert res is True
    res = task_master.add_dependent_work_units(('a', 'aa', None),
                                               ('b', 'cc', None))
    assert res is True
    # now aa can't run until bb and cc do, bb should be first
    get_and_check_unit(task_master, 'b', 'bb')
    get_and_check_unit(task_master, 'b', 'cc')
    get_and_check_unit(task_master, 'a', 'aa')
    get_and_check_done(task_master)

def test_three_failure(task_master):
    """As test_three, but the first task fails"""
    task_master.update_bundle(work_spec_a, work_units_aa, nice=0)
    task_master.update_bundle(work_spec_b, work_units_bb, nice=1)
    task_master.update_bundle(work_spec_b, work_units_cc, nice=2)
    res = task_master.add_dependent_work_units(('a', 'aa', None),
                                               ('b', 'bb', None))
    assert res is True
    res = task_master.add_dependent_work_units(('a', 'aa', None),
                                               ('b', 'cc', None))
    assert res is True

    get_and_check_unit(task_master, 'b', 'bb', fail=True)
    # a/aa should be failed right now
    assert task_master.num_failed('a') == 1
    assert task_master.num_failed('b') == 1
    assert task_master.num_available('b') == 1
    get_and_check_unit(task_master, 'b', 'cc')
    get_and_check_done(task_master)

def test_three_soft_failure(task_master):
    """As test_three_failure, but aa can still run"""
    task_master.update_bundle(work_spec_a, work_units_aa, nice=0)
    task_master.update_bundle(work_spec_b, work_units_bb, nice=1)
    task_master.update_bundle(work_spec_b, work_units_cc, nice=2)
    res = task_master.add_dependent_work_units(('a', 'aa', None),
                                               ('b', 'bb', None), hard=False)
    assert res is True
    res = task_master.add_dependent_work_units(('a', 'aa', None),
                                               ('b', 'cc', None), hard=False)
    assert res is True

    get_and_check_unit(task_master, 'b', 'bb', fail=True)
    get_and_check_unit(task_master, 'b', 'cc', fail=True)
    get_and_check_unit(task_master, 'a', 'aa')
    get_and_check_done(task_master)
    assert task_master.num_failed('a') == 0
    assert task_master.num_finished('a') == 1
    assert task_master.num_failed('b') == 2
    assert task_master.num_finished('b') == 0



''':mod:`yakonfig` declarations for rejester.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import yakonfig

config_name = 'rejester'

default_config = {
    'app_name': 'rejester',
}

def add_arguments(parser):
    parser.add_argument('--app-name',
                        help='name of app for namespace prefixing')
    parser.add_argument('--namespace',
                        help='namespace for prefixing table names')
    parser.add_argument('--registry-address', metavar='HOST:PORT',
                        action='append', dest='registry_addresses',
                        help='location of the Redis registry server')

runtime_keys = {
    'app_name': 'app_name',
    'registry_addresses': 'registry_addresses',
    'namespace': 'namespace',
}

def check_config(config, name):
    for k in ['registry_addresses', 'app_name', 'namespace']:
        if k not in config or config[k] is None:
            raise yakonfig.ConfigurationError(
                '{} requires configuration for {}'
                .format(name, k))
    if len(config['registry_addresses']) == 0:
        raise yakonfig.ConfigurationError(
            '{} requires at least one registry_addresses'
            .format(name))
    for addr in config['registry_addresses']:
        if ':' not in addr:
            raise yakonfig.ConfigurationError(
                '{} registry_addresses must be HOST:PORT, not {!r}'
                .format(name, addr))
